import copy
import time
import tkinter as tk
from PIL import Image, ImageTk, ImageOps
import matplotlib.pyplot as plt
from matplotlib import use as mpl_use
import numpy as np
import cv2
import sys
from math import log10
from datetime import datetime
import os

output_pin = 7

from picamera2 import Picamera2
import RPi.GPIO as GPIO

# Set the backend for using Tkinter with Matplotlib
mpl_use('TkAgg')

# Create the directories for all the images, files and graphs
save_path = "/home/ufyprf01/spektrofotometr"
path_to_images = save_path + "/recources"

if not os.path.exists(path_to_images + "/photos"):
    os.makedirs(path_to_images + "/photos")
if not os.path.exists(path_to_images + "/graphs"):
    os.makedirs(path_to_images + "/graphs")
if not os.path.exists(path_to_images + "/files"):
    os.makedirs(path_to_images + "/files")
if not os.path.exists(save_path + "/saves"):
    os.makedirs(save_path + "/saves")

# Set basic values such as screen size, path to images etc.
window = tk.Tk()
WIDTH = window.winfo_screenwidth()
HEIGHT = window.winfo_screenheight()
window.geometry(f"{WIDTH}x{HEIGHT}")
window.attributes('-fullscreen', True)
window.title("Spektrofotometr")
btn_text = "Avg"
img, img_2, bg, old_pic = None, None, None, None
gain_entry, mirror = False, False
run_first_time = True
fro = 1

# Defining image size by screen height so that it is still 4x3 image
img_out_size = (int(WIDTH / 4), int(HEIGHT / 3))
btn_width, btn_height = (int(img_out_size[0] / 2 - 50), int(img_out_size[1] / 8))

# Using this later we can set button size in pixels, rather than in character width
pixel = tk.PhotoImage(width=1, height=1)

GPIO.setwarnings(False)
GPIO.setmode(GPIO.BOARD)
GPIO.setup(output_pin, GPIO.OUT)
original_size = (4056, 3040)
img_ratio = (original_size[0] / img_out_size[0], original_size[1] / img_out_size[1])

picam2 = Picamera2()
picam2.configure(picam2.create_preview_configuration(raw={}))
picam2.start()



# This function is used to take the images - both .png and .raw
def get_frame(img_name):
    file = open(f"recources/photos/{img_name}.raw", 'wb')
    current_exp = exposure_time_entry.get()
    current_gain = gain_entry.get()

    picam2.set_controls({"ExposureTime": int(current_exp), "AnalogueGain": float(current_gain)})
    time.sleep(1)
    for _ in range(1):
        picam2.capture_array('raw').tofile(file)

    file.close()

    picam2.capture_file(f"recources/photos/{img_name}.jpg")
    return


# Define the function to get the first picture
def get_pic():
    # Make variables global
    global img
    GPIO.output(output_pin, True)
    get_frame("first")
    GPIO.output(output_pin, False)
    img = True
    main()


# Define the function to get the second picture
def get_2_pic():
    # Make variables global
    global img_2
    GPIO.output(output_pin, True)
    get_frame("second")
    GPIO.output(output_pin, False)
    img_2 = True
    main()



# This function is used to convert .raw format image to PIL format image
def raw_to_img(raw_img):
    # At first it converts it to a numpy array
    img_arr = np.frombuffer(raw_img, np.uint8)
    # Then it calculates how many frames are present in the .raw stream
    num_of_frames = int(img_arr.size / (1520 * 3072))
    # Then it converts the .raw stream to a 3D array - first dimension is the frame number, second is x cordinate and third is y cordinate
    img_arr = img_arr.reshape(num_of_frames, 1520, 3072)
    # Than it creates the mean of all the images and converts it to PIL image
    mean = np.mean(img_arr, axis=0)
    mean = np.rint(mean).astype(np.uint8)
    image = Image.fromarray(mean)
    return image


# This function converts the PIL image to 1D list of intensity values
def convert_bggr_to_intensity(image):
    # At first it converts it to np array and creates the mean across all y values
    img_array = np.array(image)
    img_array = np.mean(img_array, axis=0)
    blank = []
    # Then it needs to average every 3 pixels to one, because the .raw stream is 12 bits and I use 8bits here (8 + 8 + 8 bits = 24 bits = two 12 bit pixels)
    for i in range(0, img_array.shape[0] - 33, 3):
        pixel_1 = img_array[i]
        pixel_2 = img_array[i + 1]
        pixel_3 = img_array[i + 2]
        average = int((pixel_1 / 3 + pixel_2 / 3 + pixel_3 / 3))
        blank.append(average)
    return blank

# This function creates a graph from the image "image" and saves it as "im_save_name"
def plot_graph(image, im_save_name):
    # It firsts converts it to a list of intensity values
    mean = convert_bggr_to_intensity(image)

    # Than it creates the graph
    fig = plt.figure(frameon=False)
    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis()
    plt.ylabel("Light")
    plt.xlabel("Pixel")
    plt.ylim(0, 260)
    ax.plot(mean)

    # Save the graph as .png
    plt.savefig(f'{path_to_images}/graphs/{im_save_name}.png', bbox_inches='tight')
    plt.clf()
    plt.close()

    # Save the graph values so they can be accessed without opening the graph (this aproach is faster)
    f = open(f"{path_to_images}/files/just_val_{im_save_name}.txt", "w")
    for j in mean:
        try:
            j = str(int(j))
            f.write(j + "\n")
        except ValueError:
            j = str(j[0])
            f.write(j + "\n")
    f.close()


# This function is used to graph out the absorbance and transmittance
def graph_abs_and_trans():
    # First it reads the light values of the images
    text_vals_1 = open(f"{path_to_images}/files/just_val_first_img_plt.txt", "r")
    text_vals_2 = open(f"{path_to_images}/files/just_val_second_img_plt.txt", "r")

    text_vals_1 = text_vals_1.readlines()
    text_vals_2 = text_vals_2.readlines()

    # Then it calculates the absorbance and transmittance using these values
    absorbance = []
    transmittance = []
    for i, j in enumerate(text_vals_1):
        n = int(text_vals_1[i].strip("\n"))
        n2 = int(text_vals_2[i].strip("\n"))
        if n2 == 0:
            n2 = 1
        if n == 0:
            n = 1
        absorbance.append(- log10(n2 / n))
        transmittance.append((n2 / n))

    # If needed, the absorbance and transmittance are flipped
    if mirror:
        graph_absorbance = absorbance[::-1]
        graph_transmittance = transmittance[::-1]
    else:
        graph_absorbance = absorbance
        graph_transmittance = transmittance

    # And both graphs are created and saved as .png
    fig = plt.figure(frameon=False)
    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis()
    plt.ylabel("Light")
    plt.xlabel("Pixel")

    ax.plot(graph_absorbance)

    # Save the graph as .png
    plt.title("Absorbance")
    plt.savefig(f"{path_to_images}/graphs/absorbance.png", bbox_inches='tight')
    plt.clf()
    plt.close()

    fig = plt.figure(frameon=False)
    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis()
    plt.ylabel("Light")
    plt.xlabel("Pixel")

    ax.plot(graph_transmittance)

    # Save the graph as .png
    plt.title("Transmitance")
    plt.savefig(f"{path_to_images}/graphs/transmittance.png", bbox_inches='tight')
    plt.clf()
    plt.close()


# This function saves everything in the "recources" directory
def save_imgs():
    ts = name_entry.get()
    try:
        os.mkdir(f"{save_path}/saves/{ts}")
        print(f"cp -r {path_to_images} {save_path}/saves/{ts}")
        os.system(f"cp -r {path_to_images} {save_path}/saves/{ts}")
    except Exception as e:
        print(f"Saving failed at {ts}")
        print(e)



# This function stars the live video
def start_live_video():
    os.system("libcamera-hello -t 0")


# This function is used to mirror the images
def mirror_img():
    global mirror
    if mirror:
        mirror = False
    else:
        mirror = True
    main()


# Define function to end the program
def end_program():
    # GPIO.cleanup()
    sys.exit()

# This function displays the text under the images
def output_text_under_images(text: list, position: int):
    tk.Label(window, text=text[0], font=("Arial", 30)).place(x=(position + 0.3) * img_out_size[0],
                                                             y=0.5 * img_out_size[1])
    tk.Label(window, text=text[1], font=("Arial", 20)).place(x=(position + 0.3) * img_out_size[0],
                                                             y=1.5 * img_out_size[1])

# This function creates all the buttons
def buttons():
    global exposure_time_entry, gain_entry, name_entry


    # Define the button for ending the program
    end = tk.Button(window, text="Exit")
    end.config(height=btn_height, width=btn_width, image=pixel, compound="c")
    end.place(x=img_out_size[0] * 4 - btn_width - 25, y=0 + 25)
    end.configure(command=end_program)

    save_i = tk.Button(window, text="Save images")
    save_i.config(height=btn_height, width=btn_width, image=pixel, compound="c")
    save_i.place(x=img_out_size[0] * 4 - 2 * (btn_width + 25), y=2 * (btn_height + 20) + 25)
    save_i.configure(command=save_imgs)

    mirror_image = tk.Button(window, text=f"Mirror images")
    mirror_image.config(height=btn_height, width=btn_width, image=pixel, compound="c")
    mirror_image.place(x=img_out_size[0] * 4 - btn_width - 25, y=2 * (btn_height + 20) + 25)
    mirror_image.configure(command=mirror_img)

    # Define the button for taking the first picture
    getpic = tk.Button(window, text="Get 1st pic")
    getpic.config(height=btn_height, width=btn_width, image=pixel, compound="c")
    getpic.place(x=img_out_size[0] * 4 - 2 * (btn_width + 25), y=btn_height + 20 + 25)
    getpic.configure(command=get_pic)

    # Define the button for taking the second picture
    getpic_2 = tk.Button(window, text="Get 2nd pic")
    getpic_2.config(height=btn_height, width=btn_width, image=pixel, compound="c")
    getpic_2.place(x=img_out_size[0] * 4 - btn_width - 25, y=btn_height + 20 + 25)
    getpic_2.configure(command=get_2_pic)

    display_live = tk.Button(window, text="Display preview")
    display_live.config(height=btn_height, width=btn_width, image=pixel, compound="c")
    display_live.place(x=img_out_size[0] * 4 - 2 * (btn_width + 25), y=0 + 25)
    display_live.configure(command=start_live_video)


    exposure_time_entry = tk.Entry(master=window, width=6)
    exposure_time_entry.place(x=img_out_size[0] * 4 - (50), y=3 * (btn_height + 20) + 25)
    exposure_time_entry.insert(0, str(current_exp))
    tk.Label(master=window, text=f"Exposure time ({min_exp} - {max_exp})").place(
        x=img_out_size[0] * 4 - 2 * (btn_width + 25), y=3 * (btn_height + 20) + 25)

    gain_entry = tk.Entry(master=window, width=6)
    gain_entry.place(x=img_out_size[0] * 4 - (50), y=int(3.5 * (btn_height + 20)) + 25)
    gain_entry.insert(0, str(current_gain))
    tk.Label(master=window, text=f"Gain ({min_gain} - {max_gain})").place(x=img_out_size[0] * 4 - 2 * (btn_width + 25),
                                                                          y=int(3.5 * (btn_height + 20)) + 25)

    name_entry = tk.Entry(master=window, width=6)
    name_entry.place(x=img_out_size[0] * 4 - (50), y=int(4 * (btn_height + 20)) + 25)
    tk.Label(master=window, text=f"Savefile name ").place(
        x=img_out_size[0] * 4 - 2 * (btn_width + 25), y=4 * (btn_height + 20) + 25)



# This function displays all the images
def display_first_images():
    global graph_image, display_gray_slice_sub, display_org_slice_sub, display_slice_sub, tk_image
    # It first reads the images
    pil_image = Image.open(f"{path_to_images}/photos/first.jpg")
    file = open(f"{path_to_images}/photos/first.raw", "rb")
    # It converts the .raw image into PIL image
    raw_image = file.read()
    file.close()
    raw_image = raw_to_img(raw_image)

    if mirror:
        pil_image = ImageOps.mirror(pil_image)
        raw_image = ImageOps.mirror(raw_image)

    # Convert first image to Tk Image and output
    tk_image = ImageTk.PhotoImage(pil_image.resize(img_out_size))
    tk.Label(window, image=tk_image).place(x=0 * img_out_size[0], y=0 * img_out_size[1])

    # And it dispays the graph of the first image
    plot_graph(raw_image, f"first_img_plt")
    graph = Image.open(f"{path_to_images}/graphs/first_img_plt.png")
    graph_image = ImageTk.PhotoImage(graph.resize(img_out_size))
    tk.Label(window, image=graph_image).place(x=0 * img_out_size[0], y=1 * img_out_size[1])


# This function does the same thing as the previous one for the second image
def display_second_images():
    global graph_image_2, display_gray_slice_sub_2, display_org_slice_sub_2, display_slice_sub_2, tk_image_2
    pil_image = Image.open(f"{path_to_images}/photos/second.jpg")
    file = open(f"{path_to_images}/photos/second.raw", "rb")
    raw_image = file.read()
    file.close()
    raw_image = raw_to_img(raw_image)

    if mirror:
        pil_image = ImageOps.mirror(pil_image)
        raw_image = ImageOps.mirror(raw_image)

    # Convert first image to Tk Image and output
    tk_image_2 = ImageTk.PhotoImage(pil_image.resize(img_out_size))
    tk.Label(window, image=tk_image_2).place(x=1 * img_out_size[0], y=0 * img_out_size[1])


    plot_graph(raw_image, f"second_img_plt")
    graph = Image.open(f"{path_to_images}/graphs/second_img_plt.png")
    graph_image_2 = ImageTk.PhotoImage(graph.resize(img_out_size))
    tk.Label(window, image=graph_image_2).place(x=1 * img_out_size[0], y=1 * img_out_size[1])


# This function has the main logic in it, it calls all the other functions, etc.
def main(*args):
    # Add global values so that other functions can see them
    global path_to_images, label_widget, current_gain, current_exp, run_first_time, max_gain, max_exp, min_gain, min_exp

    # Define a LabelFrame that holds all the buttons
    # If the function is ran for the first time it displays the buttons and gets the values for the entries such as gain entry
    if run_first_time:
        min_exp, max_exp, current_exp = 10, 100000, 20  # picam2.camera_controls["ExposureTime"]
        min_gain, max_gain, current_gain = 1, 20, 5  # picam2.camera_controls["AnalogueGain"]
        current_gain = 3
        current_exp = 15000
        run_first_time = False
        buttons()
    # If it is not ran for the first time, it makes sure that the values such as gain and exposure are in the boundaries they are supposed to be in
    else:

        current_gain = gain_entry.get()
        current_exp = exposure_time_entry.get()

        if int(current_gain) > int(max_gain):
            current_gain = int(max_gain)
            gain_entry.delete(0, tk.END)
            gain_entry.insert(0, current_gain)
        elif int(current_gain) < int(min_gain):
            current_gain = min_gain
            gain_entry.delete(0, tk.END)
            gain_entry.insert(0, current_gain)
        if int(current_exp) > int(max_exp):
            current_exp = max_exp
            exposure_time_entry.delete(0, tk.END)
            exposure_time_entry.insert(0, current_exp)
        elif int(current_exp) < int(min_exp):
            current_exp = min_exp
            exposure_time_entry.delete(0, tk.END)
            exposure_time_entry.insert(0, current_exp)

    # It then tries to display the text under the images or the images
    if img is None or img_2 is None:
        text = ["Transmitance",
                "Absorbance"]
        output_text_under_images(text, 2)
    else:
        # If both images exist, it graphs the absorbance and transmittance
        graph_abs_and_trans()
        plot_2_bg = Image.open(f"{path_to_images}/graphs/absorbance.png")
        plot_2_out_bg = ImageTk.PhotoImage(plot_2_bg.resize(img_out_size))
        tk.Label(window, image=plot_2_out_bg).place(x=2 * img_out_size[0], y=1 * img_out_size[1])

        plot_3_bg = Image.open(f"{path_to_images}/graphs/transmittance.png")
        plot_3_out_bg = ImageTk.PhotoImage(plot_3_bg.resize(img_out_size))
        tk.Label(window, image=plot_3_out_bg).place(x=2 * img_out_size[0], y=0 * img_out_size[1])

    # This displays the first image
    if img is None:
        text = ["Image",
                "1st Graph"]
        output_text_under_images(text, 0)
    else:
        display_first_images()


    # This displays the second image
    if img_2 is None:
        text = ["Image 2",
                "2nd Graph"]
        output_text_under_images(text, 1)

    else:
        display_second_images()


    window.mainloop()


# Check if the file is opened directly and if so, run function main()
if __name__ == "__main__":
    main()
